//
//  DatabaseManager.swift
//  PowerFitnessApp
//
//  Created by user on 3/7/22.
//

import Foundation
import FirebaseFirestore
import UIKit

final class DatabaseManager {
    
    static let shared = DatabaseManager()
    
    private let database = Firestore.firestore()
    
    private init() {}
    
    public func insertUsuario(user: User, completion: @escaping (Bool) -> Void) {
        let docID = user.email.replacingOccurrences(of: ".", with: "_").replacingOccurrences(of: "@", with: "_")
        let data: [String: Any] = [
            "email": user.email,
            "nombre": user.nombre
        ]
        database.collection("users").document(docID).setData(data) { error in
            completion(error == nil)
        }
    }
    
    public func getUser(for email: String, completion: @escaping (User?) -> Void) {
        let userEmail = email.replacingOccurrences(of: ".", with: "_").replacingOccurrences(of: "@", with: "_")
        database.collection("users").document(userEmail).getDocument { snapshot, error in
            guard let data = snapshot?.data() as? [String: String], let name = data["nombre"], error == nil else {
                return
            }
            let user = User(nombre: name, email: email)
            completion(user)
        }
    }
    
    // MARK: VERSION 1.0
    
    public func getActividadGlobal(completion: @escaping ([ActividadGlobal]) -> Void) {
        database.collection("actividades").getDocuments { snapshot, error in
            guard let documents = snapshot?.documents.compactMap ({ $0.data() }), error == nil else {
                return
            }
            let actividades: [ActividadGlobal] = documents.compactMap({ dictionary in
                guard let iD = dictionary["iD"] as? String,
                      let apuntados = dictionary["apuntados"] as? Int else {
                          return nil
                      }
                let actividad = ActividadGlobal(iD: iD,
                                                apuntados: apuntados)
                return actividad
            })
            completion(actividades)
        }
    }
    
    public func getActividadUsuario(for email: String, completion: @escaping ([ActividadUsuario]) -> Void) {
        let userEmail = email.replacingOccurrences(of: ".", with: "_").replacingOccurrences(of: "@", with: "_")
        database.collection("users").document(userEmail).collection("actividades").getDocuments { snapshot, error in
            guard let documents = snapshot?.documents.compactMap ({ $0.data() }), error == nil else {
                return
            }
            let actividades: [ActividadUsuario] = documents.compactMap({ dictionary in
                guard let dia = dictionary["dia"] as? String,
                      let fecha = dictionary["fecha"] as? String,
                      let hora = dictionary["hora"] as? String,
                      let titulo = dictionary["titulo"] as? String else {
                          return nil
                      }
                let actividad = ActividadUsuario(dia: dia,
                                                 fecha: fecha,
                                                 hora: hora,
                                                 titulo: titulo)
                return actividad
            })
            completion(actividades)
        }
    }
    
    public func insertActividadUsuario(id: String, actividad: ActividadUsuario, email: String, completion: @escaping (Bool) -> Void) {
        let uEmail = email.replacingOccurrences(of: ".", with: "_").replacingOccurrences(of: "@", with: "_")
        let data: [String: Any] = [
            "dia": actividad.dia,
            "fecha": actividad.fecha,
            "hora": actividad.hora,
            "titulo": actividad.titulo
        ]
        
        database.collection("users").document(uEmail).collection("actividades").document(id).setData(data) { error in
            completion(error == nil)
        }
    }
    
    public func insertActividadGlobal(id: String, actividad: ActividadGlobal, completion: @escaping (Bool) -> Void) {
        let data: [String: Any] = [
            "iD": actividad.iD,
            "apuntados": actividad.apuntados
        ]
        database.collection("actividades").document(id).setData(data) { error in
            completion(error == nil)
        }
    }
    
    // MARK: "newNovedades"
    
    // saca las novedades que son iguales para todos los usuarios, no pasa por la coleccion de users
    public func getNovedades(completion: @escaping ([Novedad]) -> Void) {
        database.collection("novedades").getDocuments { snapshot, error in
            guard let documents = snapshot?.documents.compactMap({ $0.data() }), error == nil else {
                return
            }
            
            let novedades: [Novedad] = documents.compactMap({ dictionary in
                guard let titulo = dictionary["titulo"] as? String,
                      let imagenUrl = dictionary["imagenUrl"] as? String,
                      let id = dictionary["id"] as? String,
                      let storageUrl = dictionary["storageUrl"] as? String,
                      let cuerpo = dictionary["cuerpo"] as? String else {
                          print("...   conversion al obtener docuemntos no valida   ...")
                          return nil
                      }
                let novedad = Novedad(titulo: titulo,
                                      imagenUrl: URL(string: imagenUrl),
                                      id: id,
                                      storageUrl: storageUrl,
                                      cuerpo: cuerpo)
                return novedad
            })
            completion(novedades)
        }
    }
    
    // MARK: VERSION beta
    
    // saca las rutinas relacionadas a un usuario
    public func getRutinas(for email: String, completion: @escaping ([DocumentoPdf]) -> Void) {
        let userEmail = email.replacingOccurrences(of: ".", with: "_").replacingOccurrences(of: "@", with: "_")
        database.collection("users").document(userEmail).collection("rutinas").getDocuments { snapshot, error in
            guard let documents = snapshot?.documents.compactMap ({ $0.data() }), error == nil else {
                return
            }
            let novedades: [DocumentoPdf] = documents.compactMap({ dictionary in
                guard let id = dictionary["id"] as? String,
                      let titulo = dictionary["titulo"] as? String,
                      let docUrlString = dictionary["docUrl"] as? String,
                      let storageUrl = dictionary["storageUrl"] as? String else {
                          print("...   conversion al obtener docuemntos no valida   ...")
                          return nil
                      }
                let novedad = DocumentoPdf(titulo: titulo,
                                           docUrl: URL(string: docUrlString),
                                           id: id,
                                           storageUrl: storageUrl)
                return novedad
            })
            completion(novedades)
        }
    }
    
    // saca las dietas relacionadas a un usuario
    public func getDietas(for email: String, completion: @escaping ([DocumentoPdf]) -> Void) {
        let userEmail = email.replacingOccurrences(of: ".", with: "_").replacingOccurrences(of: "@", with: "_")
        database.collection("users").document(userEmail).collection("dietas").getDocuments { snapshot, error in
            guard let documents = snapshot?.documents.compactMap ({ $0.data() }), error == nil else {
                return
            }
            let novedades: [DocumentoPdf] = documents.compactMap({ dictionary in
                guard let id = dictionary["id"] as? String,
                      let titulo = dictionary["titulo"] as? String,
                      let docUrlString = dictionary["docUrl"] as? String,
                      let storageUrl = dictionary["storageUrl"] as? String else {
                          print("...   conversion al obtener docuemntos no valida   ...")
                          return nil
                      }
                let novedad = DocumentoPdf(titulo: titulo,
                                           docUrl: URL(string: docUrlString),
                                           id: id,
                                           storageUrl: storageUrl)
                return novedad
            })
            completion(novedades)
        }
    }
    
}


